#include <Wire.h>
#include <LiquidCrystal_I2C.h>
#include "DHT.h"

// Pin definitions
#define DHTPIN 8 // Digital pin connected to the DHT sensor
#define DHTTYPE DHT11 // DHT 11
#define SOIL_MOISTURE_SENSOR A0 // Analog pin connected to the soil moisture sensor
#define LIGHT_SENSOR A1 // Analog pin connected to the light sensor
#define PUMP 2 // Digital pin connected to the relay

// Threshold values
#define Tmin 20
#define Tmax 27
#define Lmax 80 // sun: 92
#define Mmin 40 // air: 33
#define Mmax 60 // water: 75
#define Hmax 60 // max: 100

// Variables for interrupt readings
volatile uint16_t interruptSoilMoisture;
volatile uint16_t interruptLight;

volatile unsigned long pumpStartTime = 0;
volatile bool pumpOpen = false;

bool pumpBroken = false;
uint16_t soil_moisture = 0;
uint16_t light = 0;
float humidity = 0.0;
float temperature = 0.0;
unsigned long elapsed_time = 0;

LiquidCrystal_I2C lcd(0x27, 16, 2); // LCD object with I2C address and dimensions
DHT dht(DHTPIN, DHTTYPE); // DHT sensor object

// ADC interrupt service routine for Timer1 compare match
ISR(TIMER1_COMPA_vect) {
  // Start a new ADC conversion for soil moisture
  ADMUX = (ADMUX & 0xF0) | (0x00 & 0x0F); // Set ADC input to ADC0 (pin A0)
  ADCSRA |= (1 << ADSC); // Start ADC conversion
  while (ADCSRA & (1 << ADSC)); // Wait for ADC conversion to complete
  uint16_t rawSoilMoisture = ADC;
  interruptSoilMoisture = map(rawSoilMoisture, 0, 1023, 100, 0); // Map ADC reading to soil moisture range

  // Start a new ADC conversion for light sensor
  ADMUX = (ADMUX & 0xF0) | (0x01 & 0x0F); // Set ADC input to ADC1 (pin A1)
  ADCSRA |= (1 << ADSC); // Start ADC conversion
  while (ADCSRA & (1 << ADSC)); // Wait for ADC conversion to complete
  uint16_t rawLight = ADC;
  interruptLight = map(rawLight, 0, 1023, 100, 0); // Map ADC reading to light range
}

ISR(TIMER2_COMPA_vect) {

}

// Configure Timer2
void configure_timer2() {
  TCCR2A = 0;
  TCCR2B = 0;
  TCNT2 = 0;
  OCR2A = 156; // Set compare match value for 10 ms interval (adjust as needed)
  TCCR2B |= (1 << WGM22); // Set CTC mode
  TCCR2B |= (1 << CS22) | (1 << CS21) | (1 << CS20); // Set prescaler to 1024
}

// Configure Timer1
void configure_timer1() {
  TCCR1A = 0;
  TCCR1B = 0;
  TCNT1 = 0;
  OCR1A = 62499; // Set compare match value for desired interval
  TCCR1B |= (1 << WGM12); // Set CTC mode
  TCCR1B |= (1 << CS12); // Set prescaler to 256
}

// Initialize Timer2
void init_timer2() {
  TIMSK2 |= (1 << OCIE2A); // Enable Timer2 compare match interrupt
}

// Initialize Timer1
void init_timer1() {
  TIMSK1 |= (1 << OCIE1A); // Enable Timer1 compare match interrupt
}

void setup() {
  Serial.begin(9600); // Initialize serial communication
  dht.begin(); // Initialize DHT sensor
  lcd.begin(); // Initialize LCD
  lcd.backlight(); // Turn on LCD backlight

  ADMUX = 0;
  ADMUX |= (1 << REFS0); // Set ADC reference voltage to AVCC
  ADCSRA |= (1 << ADEN); // Enable ADC
  ADCSRA |= (1 << ADPS2) | (1 << ADPS1) | (1 << ADPS0); // Set ADC prescaler to 128 for 16 MHz clock

  cli(); // Disable interrupts during timer configuration
  configure_timer1(); // Configure Timer1 for ADC readings
  init_timer1(); // Initialize Timer1
  configure_timer2(); // Configure Timer2 for pump control
  init_timer2(); // Initialize Timer2
  sei(); // Enable interrupts

  // Set pump pin as output and turn off the pump initially
  pinMode(PUMP, OUTPUT);
  digitalWrite(PUMP, HIGH);
}

// Function to start the pump
void startPump() {
  digitalWrite(PUMP, LOW); // Turn on the pump
  pumpOpen = true; // Set pumpOpen flag to indicate pump is open
  pumpStartTime = millis(); // Record the start time of the pump
  lcd.clear();
  lcd.setCursor(0, 0);
  lcd.print("Pump started!");
}

// Function to stop the pump
void stopPump() {
  digitalWrite(PUMP, HIGH); // Turn off the pump
  pumpOpen = false; // Set pumpOpen flag to indicate pump is closed
  pumpStartTime = millis(); // Record the stop time of the pump
  lcd.clear();
  lcd.setCursor(0, 0);
  lcd.print("Pump stopped!");
}

// Function to read the sensors
void readSensors() {
  // Read values from interrupt variables
  soil_moisture = interruptSoilMoisture;
  light = interruptLight;
  elapsed_time = (millis() - pumpStartTime) / 1000;

  // Read temperature and humidity from DHT sensor
  humidity = dht.readHumidity();
  temperature = dht.readTemperature();

  if (isnan(humidity) || isnan(temperature)) {
    Serial.println(F("Failed to read from temperature sensor!"));
    return;
  }
}

// Function to display sensor values and the pump's status on the LCD
void printStatus() {
  if (!pumpOpen) {
    lcd.clear();
    lcd.setCursor(0, 0);
    lcd.print("Temp: ");

    lcd.setCursor(0, 1);
    lcd.print("Humidity: ");
    lcd.setCursor(6, 0);
    lcd.print(temperature);

    lcd.setCursor(10, 1);
    lcd.print(humidity);
  } else {
    lcd.setCursor(0, 1);
    lcd.print("Time: ");
    lcd.setCursor(6, 1);
    lcd.print(elapsed_time);
    lcd.setCursor(10, 1);
    lcd.print("s");
  }

  Serial.print("Soil Moisture: ");
  Serial.println(soil_moisture);
  Serial.print("Light: ");
  Serial.println(light);
}

// Function to control the pump
void checkConditions() {
  if (soil_moisture <= Mmin) {
    if (temperature >= Tmin && temperature <= Tmax && humidity < Hmax) {
      if (light < Lmax) {
        if (!pumpOpen)
          startPump(); // Start the pump if conditions are met
      } else {
        if (pumpOpen)
          stopPump(); // Stop the pump if conditions are not met
      }
    } else {
      if (pumpOpen)
        stopPump(); // Stop the pump if conditions are not met
    }
  } else {
    if (soil_moisture >= Mmax && pumpOpen)
      stopPump(); // Stop the pump if soil moisture is above the maximum threshold
  }
}

// Function to control the pump if the sensors dont work properly
void checkPump() {
  if (pumpOpen && (millis() - pumpStartTime >= 5000)) {
    stopPump(); // Stop the pump if it has been open for over 5 seconds
    pumpBroken = true;
  }
}

void loop() {
  // Automatic pump control based on time elapsed
  if (!pumpBroken)
    checkPump();

  if (pumpBroken) {
    lcd.clear();
    lcd.setCursor(0, 0);
    lcd.print("Pump broken!");
  } else {
    // Read sensor values
    readSensors();
    
    // Display sensor values on LCD
    printStatus();
    
    // Check the conditions for pump control
    checkConditions();
  }
  
  delay(2500); // Delay between iterations
}
